/**
 * 
 */
package org.eclipse.ecf.provider.ftp.container;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import org.apache.commons.net.ftp.FTPClient;
import org.eclipse.ecf.core.ISharedObject;
import org.eclipse.ecf.core.ISharedObjectConfig;
import org.eclipse.ecf.core.ISharedObjectContext;
import org.eclipse.ecf.core.ISharedObjectManager;
import org.eclipse.ecf.core.SharedObjectInitException;
import org.eclipse.ecf.core.identity.ID;
import org.eclipse.ecf.core.util.Event;
import org.eclipse.ecf.core.util.QueueException;
import org.eclipse.ecf.fileshare.ISharedFile;
import org.eclipse.ecf.fileshare.ISharedFileEventListener;
import org.eclipse.ecf.fileshare.IStoreFileDescription;
import org.eclipse.ecf.fileshare.SharedFileCreationException;
import org.eclipse.ecf.fileshare.SharedFileStartException;

public class FtpStoreSharedObject implements ISharedObject, ISharedFile {

	public static final int DEFAULT_BUF_SIZE = 1024;
	
	FTPClient ftpclient = null;
	IStoreFileDescription spec = null;
	ISharedObjectConfig config = null;
	InputStream localInput = null;
	OutputStream remoteOutput = null;
	ISharedFileEventListener listener = null;
	boolean started = false;
	boolean cancelled = false;
	long totalSent = 0;

	public FtpStoreSharedObject(FTPClient client, IStoreFileDescription spec) throws SharedFileCreationException {
		super();
		this.ftpclient = client;
		this.spec = spec;
	}
	protected ISharedObjectContext getContext() {
		return config.getContext();
	}
	/* (non-Javadoc)
	 * @see org.eclipse.ecf.core.util.IEventHandler#handleEvent(org.eclipse.ecf.core.util.Event)
	 */
	public void handleEvent(Event event) {
		if (event instanceof StartEvent) {
			handleStartEvent();
		} else if (event instanceof SendDataEvent) {
			handleSendDataEvent();
		} else if (event instanceof EndDataEvent) {
			handleEndEvent();
		}
	}
	protected void queueSendDataEvent() {
		sendSelf(new SendDataEvent());
	}
	protected void queueEndDataEvent() {
		sendSelf(new EndDataEvent());
	}
	protected void handleStartEvent() {
		// Start sending data...
	    // XXX notify listener before starting to send data
		
		queueSendDataEvent();
	}
	protected void handleEndEvent() {
		// XXX notify listener about end successfully completed
		destroySelfLocal();
	}
	protected void handleIOException(IOException e) {
		// XXX notify listener of problem, then kill ourselves
		
		destroySelfLocal();
	}
	protected void handleSendDataEvent() {
		if (cancelled) {
			// XXX first notify listener that the send has been cancelled by user
			destroySelfLocal();
			return;
		} else {
			try {
				byte [] buf = new byte[DEFAULT_BUF_SIZE];
				int read = localInput.read(buf);
				if (read == -1) {
					queueEndDataEvent();
					return;
				}
				// XXX notify listener here about pre write
				remoteOutput.write(buf,0,read);
				// XXX notify listner here about post write
				totalSent += read;
			} catch (IOException e) {
				handleIOException(e);
				return;
			}
			queueSendDataEvent();
		}
	}
	/* (non-Javadoc)
	 * @see org.eclipse.ecf.core.util.IEventHandler#handleEvents(org.eclipse.ecf.core.util.Event[])
	 */
	public void handleEvents(Event[] events) {
		for(int i=0; i < events.length; i++) {
			handleEvent(events[i]);
		}
	}
	
	// ISharedFile
	public synchronized void start() throws SharedFileStartException {
		try {
			this.localInput = spec.getLocalStream();
		} catch (IOException e) {
			throw new SharedFileStartException("cound not get local stream",e);
		}
		if (localInput == null) throw new SharedFileStartException("input stream cannot be null");
		listener = spec.getEventListener();
		String remoteFileName = spec.getRemoteFileName();
		try {
			remoteOutput = ftpclient.storeFileStream(remoteFileName);
		} catch (IOException e) {
			throw new SharedFileStartException("exception storing remote file with name "+remoteFileName,e);
		}
		started = true;
		cancelled = false;
		sendSelf(new StartEvent());
	}
	public synchronized void cancel() {
		cancelled = true;
	}
    protected void destroySelfLocal() {
        try {
            ISharedObjectConfig soconfig = config;
            if (soconfig != null) {
                ID myID = soconfig.getSharedObjectID();
                ISharedObjectContext context = getContext();
                if (context != null) {
                    ISharedObjectManager manager = context.getSharedObjectManager();
                    if (manager != null) {
                        manager.removeSharedObject(myID);
                    }
                }
            }
        } catch (Exception e) {
            //traceDump("Exception in destroySelfLocal()",e);
        }
    }
	/* (non-Javadoc)
	 * @see org.eclipse.ecf.core.ISharedObject#init(org.eclipse.ecf.core.ISharedObjectConfig)
	 */
	public void init(ISharedObjectConfig initData)
			throws SharedObjectInitException {
		this.config = initData;
	}
	/* (non-Javadoc)
	 * @see org.eclipse.ecf.core.ISharedObject#dispose(org.eclipse.ecf.core.identity.ID)
	 */
	public void dispose(ID containerID) {
		// we're doomed, simply clean up
		try {
			if (localInput != null) {
				localInput.close();
				localInput = null;
			}
		} catch (IOException e) {}
		try {
			if (remoteOutput != null) {
				remoteOutput.close();
				remoteOutput = null;
			}
		} catch (IOException e) {}
		listener = null;
		config = null;
		spec = null;
		ftpclient = null;
	}
	/* (non-Javadoc)
	 * @see org.eclipse.core.runtime.IAdaptable#getAdapter(java.lang.Class)
	 */
	public Object getAdapter(Class adapter) {
		return null;
	}
	class StartEvent implements Event {
		private static final long serialVersionUID = -3494636993772750229L;
	}
	class SendDataEvent implements Event {
		private static final long serialVersionUID = 3058901983918082320L;
	}
	class EndDataEvent implements Event {
		private static final long serialVersionUID = -4749385115513108904L;
	}
	protected void sendSelf(Event evt) {
		try {
			getContext().getQueue().enqueue(new SendDataEvent());
		} catch (QueueException e) {
			// XXX need to report this to debug output, etc
			e.printStackTrace();
		}
	}

}
